#include "GOAPSolver.h"
#include "CMacros.h"
#include "CWindow.h"
#include "GOAPAction.h"

using namespace std;
using namespace Combustion;

namespace Game
{
	GOAPSolver::GOAPSolver(GOAPState& e, map<int, string>& idnames, vector<GOAPAction>& actions)
		: end(e), idtoname(idnames), actionSet(actions), foundEnd(false)
	{}

	GOAPSolver::~GOAPSolver() 
	{
		ClearStates();
	}

	void GOAPSolver::ClearStates()
	{
		for(auto it = generatedStates.begin(); it != generatedStates.end(); ++it)
			delete (*it);
		generatedStates.clear();
	}

	void GOAPSolver::SetStart(GOAPState& startState, GOAPResources& resources)
	{ 
		foundEnd = false;
		
		start = GOAPState(startState);
		start.SetWorldState(resources);
		current = start;
		open.clear();
		
		ClearStates();
		Open(&current);
	}

	GOAPState* GOAPSolver::GetCheapest()
	{
		GOAPState* state = *(open.begin());
		return state;
	}

	void GOAPSolver::Open(GOAPState* state)
	{
		state->SetIsClosed(false);
		state->SetIsOpen(true);
		open.insert(state);
	}

	void GOAPSolver::Close(GOAPState* state)
	{
		state->SetIsClosed(true);
		state->SetIsOpen(false);
	}

	void GOAPSolver::RemoveOpen()
	{
		open.erase( open.begin() );
	}

	bool GOAPSolver::ReachedEnd(GOAPState* state)
	{
		return state->IsSatisfied();
	}

	void GOAPSolver::GeneratePath(GOAPState* state, std::vector<GOAPAction>& list)
	{
		if(state->GetAction() != 0)
		{
			list.push_back(*(state->GetAction()));
		}

		if(state->GetParent() != 0)
			GeneratePath(state->GetParent(), list);
	}

	void GOAPSolver::PerformOneStep(std::vector<GOAPAction>& actionList)
	{
		if(open.size() == 0)
			return;

		GOAPState* currstate = GetCheapest();

		if( ReachedEnd(currstate) )
		{
			foundEnd = true;
			GeneratePath(currstate, actionList);
			return;
		}

		//move current to closed set
		RemoveOpen();
		Close(currstate);

		//find all actions that can be performed on the current state
		for(unsigned int i = 0; i < actionSet.size(); ++i)
		{
			if( actionSet[i].MeetsPrecondition(*currstate) )
			{
				//Get state if this action is performed
				GOAPState* neighbor = actionSet[i].ApplyToState(*currstate);
				neighbor->SetAction(&(actionSet[i]));

				if(!isInGenerated(neighbor))
				{
					neighbor->SetIsOpen(true);
					AddToGenerated(neighbor);
				}
				else
				{
					GOAPState* temp = GetGenerated(neighbor);
					delete neighbor;
					if(neighbor->GetIsClosed())
						continue;
					neighbor = temp;
				}

				//if neighbor isOpen
				if(neighbor->GetIsOpen())
				{
					neighbor->SetGVal( currstate->GetGVal() + actionSet[i].GetGCost() );
					neighbor->SetHVal( ComputeHeuristic(neighbor) );
					neighbor->SetParent(currstate);

					//check to see if the path to this one is better....
					if(currstate->GetGVal() < neighbor->GetGVal())
					{
						neighbor->SetParent(currstate);
						neighbor->SetGVal(currstate->GetGVal() + actionSet[i].GetGCost());
					}

					Open(neighbor);
				}
			}
		}
	}

	void GOAPSolver::PerformAStar(std::vector<GOAPAction>& actionList)
	{
		int maxAttempts = 10000;

		while(!foundEnd && maxAttempts > 0)
		{
			PerformOneStep(actionList);
			maxAttempts--;
		}
	}

	float GOAPSolver::ComputeHeuristic(GOAPState* state)
	{
		return end.ComputeNumDifferences(*state);
	}

	bool GOAPSolver::isInGenerated(GOAPState* s)
	{
		return (generatedStates.find(s) != generatedStates.end());
	}

	GOAPState* GOAPSolver::GetGenerated(GOAPState* s)
	{
		return *(generatedStates.find(s));
	}

	void GOAPSolver::AddToGenerated(GOAPState* s)
	{
		generatedStates.insert(s);
	}
}