#pragma once

#include <map>
#include <vector>
#include <string>
#include <set>
#include "GOAPAction.h"
#include "GOAPState.h"
#include "GOAPResources.h"

namespace Game
{
	/*
		Comparision function for states
		based on the resources of each state
	*/
	struct StateComp
	{
		bool operator()(GOAPState* a, GOAPState* b)
		{
			return a->Comp(*b);
		}
	};

	/*
		Used to sort GOAP states in the state list
		based on the HValues
	*/
	struct GOAPComp
	{
		bool operator()(GOAPState* a, GOAPState* b)
		{
			if(a->GetScore() == b->GetScore())
			{
				if(a->GetHVal() == b->GetHVal())
					return a < b;
				else
					return a->GetHVal() < b->GetHVal();
			}
			else
				return a->GetScore() < b->GetScore();
		}
	};


	class GOAPSolver
	{
	public:
		GOAPSolver(GOAPState& end, std::map<int, std::string>& idnames,
					std::vector<GOAPAction>& actions);
		~GOAPSolver();

		/*
			Sets the starting state for the solver, and the resources
			owned by the character
		*/
		void SetStart(GOAPState& startState, GOAPResources& presources);
		
		/*
			Searches all possibilities to find the best course of action
			(Limited to 10,000 attempts)
		*/
		void PerformAStar(std::vector<GOAPAction>& actionList);

	private:
		std::map<int, std::string> idtoname;//used for printing plans
		std::vector<GOAPAction> actionSet;
		std::set<GOAPState*, StateComp> generatedStates;
		std::set<GOAPState*, GOAPComp> open;

		GOAPState start;
		GOAPState end;
		GOAPState current;
		bool foundEnd;

		/*
			Checks the "neighboring" states to find the possible actions
			that can be performed (with their outcomes)
			and computes the HVal and GVal for those actions
		*/
		void PerformOneStep(std::vector<GOAPAction>& actionList);
		
				/*
			Returns the cheapest option from the open list
		*/
		GOAPState* GetCheapest();
		
		/*
			Removes the cheapest option from the open list
		*/
		void RemoveOpen();
		
		/*
			Adds the state to the open list
		*/
		void Open(GOAPState* state);
		
		/*
			Removes the state from the open list and adds it to the closed list
		*/
		void Close(GOAPState* state);

		/*
			Checks to see if the state satisfies all of the end conditions
		*/
		bool ReachedEnd(GOAPState* state);
		
		/*
			Creates the list of actions needed to get from the start state
			to the end state
		*/
		void GeneratePath(GOAPState* state, std::vector<GOAPAction>& list);

		/*
			Uses the number of differences between the current state and the given
			state as the heuristic 
			(more differences are worse)
		*/
		float ComputeHeuristic(GOAPState* state);

		/*
			Checks to see if the given state can already be reached through
			another method (that is, already exists in the list)
		*/
		bool isInGenerated(GOAPState* s);
		
		/*
			Returns the state from the list of computed states
		*/
		GOAPState* GetGenerated(GOAPState* s);
		
		/*
			Adds the state to the list of computed states
		*/
		void AddToGenerated(GOAPState* s);
		
		/*
			Clear all computed states
		*/
		void ClearStates();
	};
}